/*  *************************************************************  */
/*                                                                 */
/*  This program is a MPI version of the code to find integration  */
/*  of a given function. Here P0 is the master processor and P1    */
/*  to P(N-1) processors are slave.                                */
/*                                                                 */
/*  Author: Samir Mehta, JSU for CPS615 Prog-Assignment # 1        */
/*  Date  : 2/15/2000          Version  : 1.0                      */
/*                                                                 */
/*  *************************************************************  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

/* double f(double); */

main(int argc, char **argv)
{
  int          rank, size, tag, ierr, start, end, chunks;
  MPI_Status   status;
  double       pintegral;

  double       deltax, pi = 3.141592654, integral, fleftx, 
               frightx, rightx, leftx, f();
  int          nintervals = 128, i;
  double t1, t2;

  
  ierr = MPI_Init(&argc, &argv);
  ierr = MPI_Comm_size(MPI_COMM_WORLD, &size);
  ierr = MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  t1 = MPI_Wtime();

  /* Dividing the nintervals into small chunks for each processors */
  /* If we have 4 processors assigned then chunks = 32             */
  chunks    = nintervals/size;

  /* So that start will be 1, 33, 65, 97 for P0, P1, P2, P3 resp.  */
  start     = rank * chunks + 1;

  /* end will be 32, 64, 96, 128 for P0, P1, P2, P3 resp.          */
  end       = (rank + 1) * chunks;

  deltax    = (0.5*pi)/(double)nintervals;
  integral  = 0.0;
  leftx     = deltax * (double)(start - 1);
  fleftx    = f(leftx);

  for (i=start; i<=end; i++)
  {
    rightx    = deltax * (double) i;
    frightx   = f(rightx);
    integral += deltax * 0.5 *(fleftx + frightx);
    fleftx    = frightx;
  }
  printf ("Processor %d: contribution = %f\n", rank, integral);
  /* Common code for all processors ends here                      */

  /* Processor specific code starts here                           */
  tag       = 100;
  if (rank == 0)  /* Code for master processor                     */
  {
    for (i=1; i<size; i++)
    {
      ierr = MPI_Recv(&pintegral, 1, MPI_DOUBLE, i, tag, MPI_COMM_WORLD, &status);
      integral += pintegral;
    }
    t2 = MPI_Wtime();
    printf ("\n Integral = %f\n Elapse time is %f seconds\n", integral, (t2-t1));
  }
  else            /* Code for slave processor/s                    */
  {
    ierr = MPI_Send(&integral, 1, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD);
  }
  
  ierr = MPI_Finalize();
  exit(0);
}

double f(double x)
{
  double y;
  y = cos(x);
  return (1.0/sqrt(1.0 - 0.5*y*y));
}